#include "traceStream.hpp"
#include "../trace.h"
#include <Chronos.hpp>

#define MAX_LINE_LEN 128

#ifdef TRACE_LOG_COLORS
    #define TRACE_LOG_COLOR_BLACK         "30"
    #define TRACE_LOG_COLOR_RED           "31" // ERROR
    #define TRACE_LOG_COLOR_GREEN         "32" // INFO
    #define TRACE_LOG_COLOR_YELLOW        "33" // WARNING
    #define TRACE_LOG_COLOR_BLUE          "34"
    #define TRACE_LOG_COLOR_MAGENTA       "35"
    #define TRACE_LOG_COLOR_CYAN          "36" // DEBUG
    #define TRACE_LOG_COLOR_GRAY          "37" // VERBOSE
    #define TRACE_LOG_COLOR_WHITE         "38"

    #define TRACE_LOG_COLOR(COLOR)        "\033[0;" COLOR "m"
    #define TRACE_LOG_BOLD(COLOR)         "\033[1;" COLOR "m"
    #define TRACE_LOG_RESET_COLOR         "\033[0m"

    #define TRACE_LOG_COLOR_E             TRACE_LOG_COLOR(TRACE_LOG_COLOR_RED)
    #define TRACE_LOG_COLOR_W             TRACE_LOG_COLOR(TRACE_LOG_COLOR_YELLOW)
    #define TRACE_LOG_COLOR_I             TRACE_LOG_COLOR(TRACE_LOG_COLOR_GREEN)
    #define TRACE_LOG_COLOR_D             TRACE_LOG_COLOR(TRACE_LOG_COLOR_CYAN)
    #define TRACE_LOG_COLOR_V             TRACE_LOG_COLOR(TRACE_LOG_COLOR_GRAY)
    #define TRACE_LOG_COLOR_PRINT(letter) log_printf(TRACE_LOG_COLOR_##letter)
    #define TRACE_LOG_COLOR_PRINT_END     log_printf(TRACE_LOG_RESET_COLOR)
#else
    #define TRACE_LOG_COLOR_E
    #define TRACE_LOG_COLOR_W
    #define TRACE_LOG_COLOR_I
    #define TRACE_LOG_COLOR_D
    #define TRACE_LOG_COLOR_V
    #define TRACE_LOG_RESET_COLOR
    #define TRACE_LOG_COLOR_PRINT(letter)
    #define TRACE_LOG_COLOR_PRINT_END
#endif

#define TRACE_SHORT_LOG_FORMAT(letter, format) PSTR(TRACE_LOG_COLOR_##letter "%s" TRACE_LOG_RESET_COLOR "\r\n"), format
#define TRACE_LOG_FORMAT(letter, format)       PSTR(TRACE_LOG_COLOR_##letter "[%8lu][" #letter "] %s" TRACE_LOG_RESET_COLOR "\r\n"), millis(), format

static void (*message_cb)(const char *) = nullptr;

void TraceStream::print(const uint8_t severity, const char * buffer, const int length) {

    if (mpOut) {
        if (severity == TRACE_ERROR)
            mpOut->printf_P(TRACE_LOG_FORMAT(E, buffer));
        else if (severity == TRACE_DEBUG)
            mpOut->printf_P(TRACE_LOG_FORMAT(D, buffer));
        else if (severity == TRACE_WARNING)
            mpOut->printf_P(TRACE_LOG_FORMAT(W, buffer));
        else if (severity == TRACE_INFO)
            mpOut->printf_P(TRACE_LOG_FORMAT(I, buffer));
        else if (severity == TRACE_VERBOSE)
            mpOut->printf_P(TRACE_LOG_FORMAT(V, buffer));
    }
}

void TraceStream::init() {
}

void TraceStream::print(const uint8_t severity, const __FlashStringHelper * fmt, ...) {
    char buffer[MAX_LINE_LEN + 1];
    va_list args;

    va_start(args, fmt);
    const int length = vsnprintf_P(buffer, sizeof(buffer), (const char *) fmt, args);
    va_end(args);

    print(severity, buffer, length);
}

void TraceStream::print(const uint8_t severity, const char * fmt, ...) {
    char buffer[MAX_LINE_LEN + 1];
    va_list args;

    va_start(args, fmt);
    const int length = vsnprintf(buffer, sizeof(buffer), fmt, args);
    va_end(args);

    print(severity, buffer, length);
}

void trace_registermessagecb(void (*cb)(const char *)) {

    message_cb = cb;
}

char TraceStream::hexascii(uint8_t n) {

    n &= 0xf;
    if (n > 9)
        return n + ('A' - 10);
    else
        return n + '0';
}

void TraceStream::dump(const uint8_t severity, const char * prefix, const uint8_t * address, size_t size) {
    char buffer[MAX_LINE_LEN + 1 + 3];
    int idx = 0;

    if (prefix)
        idx = snprintf_P(buffer, MAX_LINE_LEN, PSTR("%s"), prefix);

    while ((idx < MAX_LINE_LEN) && size) {
        buffer[idx] = hexascii(*address >> 4);
        ++idx;
        buffer[idx] = hexascii(*address);
        ++idx;
        buffer[idx] = 0x20;
        ++idx;
        ++address;
        --size;
    }
    buffer[idx] = 0;
    print(severity, buffer, idx);
}

void TraceStream::dump(const uint8_t severity, const __FlashStringHelper * prefix, const uint8_t * address, size_t size) {
    char buffer[MAX_LINE_LEN + 1 + 3];
    int idx = 0;

    if (prefix) {
        char dummy[64];

        strcpy_P(dummy, (const char *) prefix);
        idx = snprintf_P(buffer, MAX_LINE_LEN, PSTR("%s"), dummy);
    }

    while ((idx < MAX_LINE_LEN) && size) {
        buffer[idx] = hexascii(*address >> 4);
        ++idx;
        buffer[idx] = hexascii(*address);
        ++idx;
        buffer[idx] = 0x20;
        ++idx;
        ++address;
        --size;
    }
    buffer[idx] = 0;
    print(severity, buffer, idx);
}

// EOF
